"""
Class information for DataFrame iterators returned by pd.read_csv. This is used
to handle situations in which pd.read_csv is used to return chunks with separate
read calls instead of just a single read.
"""
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, ir_utils, types
from numba.core.imputils import RefType, impl_ret_borrowed, iternext_impl
from numba.core.typing.templates import signature
from numba.extending import intrinsic, lower_builtin, models, register_model
import bodo
import bodo.ir.connector
import bodo.ir.csv_ext
from bodo import objmode
from bodo.hiframes.pd_dataframe_ext import DataFrameType
from bodo.hiframes.table import Table, TableType
from bodo.io import csv_cpp
from bodo.ir.csv_ext import _gen_read_csv_objmode, astype
from bodo.utils.typing import ColNamesMetaType
from bodo.utils.utils import check_java_installation
from bodo.utils.utils import sanitize_varname
ll.add_symbol('update_csv_reader', csv_cpp.update_csv_reader)
ll.add_symbol('initialize_csv_reader', csv_cpp.initialize_csv_reader)


class CSVIteratorType(types.SimpleIteratorType):

    def __init__(self, df_type, out_colnames, out_types, usecols, sep,
        index_ind, index_arr_typ, index_name, escapechar, storage_options):
        assert isinstance(df_type, DataFrameType
            ), 'CSVIterator must return a DataFrame'
        wezgu__cqd = (
            f'CSVIteratorType({df_type}, {out_colnames}, {out_types}, {usecols}, {sep}, {index_ind}, {index_arr_typ}, {index_name}, {escapechar})'
            )
        super(types.SimpleIteratorType, self).__init__(wezgu__cqd)
        self._yield_type = df_type
        self._out_colnames = out_colnames
        self._out_types = out_types
        self._usecols = usecols
        self._sep = sep
        self._index_ind = index_ind
        self._index_arr_typ = index_arr_typ
        self._index_name = index_name
        self._escapechar = escapechar
        self._storage_options = storage_options

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(CSVIteratorType)
class CSVIteratorModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        tun__ktm = [('csv_reader', bodo.ir.connector.stream_reader_type), (
            'index', types.EphemeralPointer(types.uintp))]
        super(CSVIteratorModel, self).__init__(dmm, fe_type, tun__ktm)


@lower_builtin('getiter', CSVIteratorType)
def getiter_csv_iterator(context, builder, sig, args):
    uekmj__xnoq = cgutils.create_struct_proxy(sig.args[0])(context, builder,
        value=args[0])
    txbm__qgsse = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
        as_pointer()])
    yxg__ustdf = cgutils.get_or_insert_function(builder.module, txbm__qgsse,
        name='initialize_csv_reader')
    builder.call(yxg__ustdf, [uekmj__xnoq.csv_reader])
    builder.store(context.get_constant(types.uint64, 0), uekmj__xnoq.index)
    return impl_ret_borrowed(context, builder, sig.return_type, args[0])


@lower_builtin('iternext', CSVIteratorType)
@iternext_impl(RefType.NEW)
def iternext_csv_iterator(context, builder, sig, args, result):
    [efbd__pmmnk] = sig.args
    [cuuic__uxwnz] = args
    uekmj__xnoq = cgutils.create_struct_proxy(efbd__pmmnk)(context, builder,
        value=cuuic__uxwnz)
    txbm__qgsse = lir.FunctionType(lir.IntType(1), [lir.IntType(8).
        as_pointer()])
    yxg__ustdf = cgutils.get_or_insert_function(builder.module, txbm__qgsse,
        name='update_csv_reader')
    gymjf__jxu = builder.call(yxg__ustdf, [uekmj__xnoq.csv_reader])
    result.set_valid(gymjf__jxu)
    with builder.if_then(gymjf__jxu):
        ecogv__cbad = builder.load(uekmj__xnoq.index)
        dta__ikriu = types.Tuple([sig.return_type.first_type, types.int64])
        broqo__aur = gen_read_csv_objmode(sig.args[0])
        lgto__dagw = signature(dta__ikriu, bodo.ir.connector.
            stream_reader_type, types.int64)
        mog__zjkpc = context.compile_internal(builder, broqo__aur,
            lgto__dagw, [uekmj__xnoq.csv_reader, ecogv__cbad])
        dxmq__wcwmv, uape__gys = cgutils.unpack_tuple(builder, mog__zjkpc)
        xfnaz__qzf = builder.add(ecogv__cbad, uape__gys, flags=['nsw'])
        builder.store(xfnaz__qzf, uekmj__xnoq.index)
        result.yield_(dxmq__wcwmv)


@intrinsic
def init_csv_iterator(typingctx, csv_reader, csv_iterator_typeref):

    def codegen(context, builder, signature, args):
        ujdo__ajut = cgutils.create_struct_proxy(signature.return_type)(context
            , builder)
        ujdo__ajut.csv_reader = args[0]
        oeyc__dfz = context.get_constant(types.uintp, 0)
        ujdo__ajut.index = cgutils.alloca_once_value(builder, oeyc__dfz)
        return ujdo__ajut._getvalue()
    assert isinstance(csv_iterator_typeref, types.TypeRef
        ), 'Initializing a csv iterator requires a typeref'
    odf__mjsl = csv_iterator_typeref.instance_type
    sig = signature(odf__mjsl, csv_reader, csv_iterator_typeref)
    return sig, codegen


def gen_read_csv_objmode(csv_iterator_type):
    bmh__ksqm = 'def read_csv_objmode(f_reader):\n'
    gdt__mmvwo = [sanitize_varname(ezmrf__nkk) for ezmrf__nkk in
        csv_iterator_type._out_colnames]
    wqtoz__glf = ir_utils.next_label()
    rzvz__gkgz = globals()
    out_types = csv_iterator_type._out_types
    rzvz__gkgz[f'table_type_{wqtoz__glf}'] = TableType(tuple(out_types))
    rzvz__gkgz[f'idx_array_typ'] = csv_iterator_type._index_arr_typ
    ulbu__aufi = list(range(len(csv_iterator_type._usecols)))
    bmh__ksqm += _gen_read_csv_objmode(csv_iterator_type._out_colnames,
        gdt__mmvwo, out_types, csv_iterator_type._usecols, ulbu__aufi,
        csv_iterator_type._sep, csv_iterator_type._escapechar,
        csv_iterator_type._storage_options, wqtoz__glf, rzvz__gkgz,
        parallel=False, check_parallel_runtime=True, idx_col_index=
        csv_iterator_type._index_ind, idx_col_typ=csv_iterator_type.
        _index_arr_typ)
    sfi__akebs = bodo.ir.csv_ext._gen_parallel_flag_name(gdt__mmvwo)
    lclui__jlvh = ['T'] + (['idx_arr'] if csv_iterator_type._index_ind is not
        None else []) + [sfi__akebs]
    bmh__ksqm += f"  return {', '.join(lclui__jlvh)}"
    rzvz__gkgz = globals()
    pbzn__unye = {}
    exec(bmh__ksqm, rzvz__gkgz, pbzn__unye)
    dpl__qryn = pbzn__unye['read_csv_objmode']
    mwqn__bfoc = numba.njit(dpl__qryn)
    bodo.ir.csv_ext.compiled_funcs.append(mwqn__bfoc)
    rbj__fqdr = 'def read_func(reader, local_start):\n'
    rbj__fqdr += f"  {', '.join(lclui__jlvh)} = objmode_func(reader)\n"
    index_ind = csv_iterator_type._index_ind
    if index_ind is None:
        rbj__fqdr += f'  local_len = len(T)\n'
        rbj__fqdr += '  total_size = local_len\n'
        rbj__fqdr += f'  if ({sfi__akebs}):\n'
        rbj__fqdr += """    local_start = local_start + bodo.libs.distributed_api.dist_exscan(local_len, _op)
"""
        rbj__fqdr += (
            '    total_size = bodo.libs.distributed_api.dist_reduce(local_len, _op)\n'
            )
        nnk__fmec = (
            f'bodo.hiframes.pd_index_ext.init_range_index(local_start, local_start + local_len, 1, None)'
            )
    else:
        rbj__fqdr += '  total_size = 0\n'
        nnk__fmec = (
            f'bodo.utils.conversion.convert_to_index({lclui__jlvh[1]}, {csv_iterator_type._index_name!r})'
            )
    rbj__fqdr += f"""  return (bodo.hiframes.pd_dataframe_ext.init_dataframe(({lclui__jlvh[0]},), {nnk__fmec}, __col_name_meta_value_read_csv_objmode), total_size)
"""
    exec(rbj__fqdr, {'bodo': bodo, 'objmode_func': mwqn__bfoc, '_op': np.
        int32(bodo.libs.distributed_api.Reduce_Type.Sum.value),
        '__col_name_meta_value_read_csv_objmode': ColNamesMetaType(
        csv_iterator_type.yield_type.columns)}, pbzn__unye)
    return pbzn__unye['read_func']
