import numba
import numpy as np
import pandas as pd
from numba.core import ir, ir_utils, typeinfer, types
from numba.core.ir_utils import compile_to_numba_ir, replace_arg_nodes
import bodo
import bodo.ir.connector
from bodo import objmode
from bodo.io.fs_io import get_storage_options_pyobject, storage_options_dict_type
from bodo.libs.str_ext import string_type
from bodo.transforms import distributed_analysis, distributed_pass
from bodo.utils.utils import check_java_installation, sanitize_varname


class JsonReader(ir.Stmt):

    def __init__(self, df_out, loc, out_vars, out_types, file_name,
        df_colnames, orient, convert_dates, precise_float, lines,
        compression, storage_options):
        self.connector_typ = 'json'
        self.df_out = df_out
        self.loc = loc
        self.out_vars = out_vars
        self.out_types = out_types
        self.file_name = file_name
        self.df_colnames = df_colnames
        self.orient = orient
        self.convert_dates = convert_dates
        self.precise_float = precise_float
        self.lines = lines
        self.compression = compression
        self.storage_options = storage_options

    def __repr__(self):
        return ('{} = ReadJson(file={}, col_names={}, types={}, vars={})'.
            format(self.df_out, self.file_name, self.df_colnames, self.
            out_types, self.out_vars))


import llvmlite.binding as ll
from bodo.io import json_cpp
ll.add_symbol('json_file_chunk_reader', json_cpp.json_file_chunk_reader)
json_file_chunk_reader = types.ExternalFunction('json_file_chunk_reader',
    bodo.ir.connector.stream_reader_type(types.voidptr, types.bool_, types.
    bool_, types.int64, types.voidptr, types.voidptr,
    storage_options_dict_type))


def remove_dead_json(json_node, lives_no_aliases, lives, arg_aliases,
    alias_map, func_ir, typemap):
    kok__yyiy = []
    kfkba__agv = []
    qmkdd__xfyzq = []
    for fyaqa__cac, pmkw__wqx in enumerate(json_node.out_vars):
        if pmkw__wqx.name in lives:
            kok__yyiy.append(json_node.df_colnames[fyaqa__cac])
            kfkba__agv.append(json_node.out_vars[fyaqa__cac])
            qmkdd__xfyzq.append(json_node.out_types[fyaqa__cac])
    json_node.df_colnames = kok__yyiy
    json_node.out_vars = kfkba__agv
    json_node.out_types = qmkdd__xfyzq
    if len(json_node.out_vars) == 0:
        return None
    return json_node


def json_distributed_run(json_node, array_dists, typemap, calltypes,
    typingctx, targetctx):
    if bodo.user_logging.get_verbose_level() >= 1:
        qamue__skcx = (
            'Finish column pruning on read_json node:\n%s\nColumns loaded %s\n'
            )
        hiuy__xlc = json_node.loc.strformat()
        vdvgw__gusny = json_node.df_colnames
        bodo.user_logging.log_message('Column Pruning', qamue__skcx,
            hiuy__xlc, vdvgw__gusny)
        vczym__jsv = [zlzt__xjpt for fyaqa__cac, zlzt__xjpt in enumerate(
            json_node.df_colnames) if isinstance(json_node.out_types[
            fyaqa__cac], bodo.libs.dict_arr_ext.DictionaryArrayType)]
        if vczym__jsv:
            meyyq__zht = """Finished optimized encoding on read_json node:
%s
Columns %s using dictionary encoding to reduce memory usage.
"""
            bodo.user_logging.log_message('Dictionary Encoding', meyyq__zht,
                hiuy__xlc, vczym__jsv)
    parallel = False
    if array_dists is not None:
        parallel = True
        for ljeo__rhrut in json_node.out_vars:
            if array_dists[ljeo__rhrut.name
                ] != distributed_pass.Distribution.OneD and array_dists[
                ljeo__rhrut.name] != distributed_pass.Distribution.OneD_Var:
                parallel = False
    pdfip__clcvk = len(json_node.out_vars)
    kkqce__dbox = ', '.join('arr' + str(fyaqa__cac) for fyaqa__cac in range
        (pdfip__clcvk))
    zkc__mzijf = 'def json_impl(fname):\n'
    zkc__mzijf += '    ({},) = _json_reader_py(fname)\n'.format(kkqce__dbox)
    nmsn__fam = {}
    exec(zkc__mzijf, {}, nmsn__fam)
    qny__rqvqq = nmsn__fam['json_impl']
    cyyr__kkfg = _gen_json_reader_py(json_node.df_colnames, json_node.
        out_types, typingctx, targetctx, parallel, json_node.orient,
        json_node.convert_dates, json_node.precise_float, json_node.lines,
        json_node.compression, json_node.storage_options)
    wnsko__zbcxk = compile_to_numba_ir(qny__rqvqq, {'_json_reader_py':
        cyyr__kkfg}, typingctx=typingctx, targetctx=targetctx, arg_typs=(
        string_type,), typemap=typemap, calltypes=calltypes).blocks.popitem()[1
        ]
    replace_arg_nodes(wnsko__zbcxk, [json_node.file_name])
    juzth__kuz = wnsko__zbcxk.body[:-3]
    for fyaqa__cac in range(len(json_node.out_vars)):
        juzth__kuz[-len(json_node.out_vars) + fyaqa__cac
            ].target = json_node.out_vars[fyaqa__cac]
    return juzth__kuz


numba.parfors.array_analysis.array_analysis_extensions[JsonReader
    ] = bodo.ir.connector.connector_array_analysis
distributed_analysis.distributed_analysis_extensions[JsonReader
    ] = bodo.ir.connector.connector_distributed_analysis
typeinfer.typeinfer_extensions[JsonReader
    ] = bodo.ir.connector.connector_typeinfer
ir_utils.visit_vars_extensions[JsonReader
    ] = bodo.ir.connector.visit_vars_connector
ir_utils.remove_dead_extensions[JsonReader] = remove_dead_json
numba.core.analysis.ir_extension_usedefs[JsonReader
    ] = bodo.ir.connector.connector_usedefs
ir_utils.copy_propagate_extensions[JsonReader
    ] = bodo.ir.connector.get_copies_connector
ir_utils.apply_copy_propagate_extensions[JsonReader
    ] = bodo.ir.connector.apply_copies_connector
ir_utils.build_defs_extensions[JsonReader
    ] = bodo.ir.connector.build_connector_definitions
distributed_pass.distributed_run_extensions[JsonReader] = json_distributed_run
compiled_funcs = []


def _gen_json_reader_py(col_names, col_typs, typingctx, targetctx, parallel,
    orient, convert_dates, precise_float, lines, compression, storage_options):
    pkg__mrc = [sanitize_varname(zlzt__xjpt) for zlzt__xjpt in col_names]
    abrz__gyvck = ', '.join(str(fyaqa__cac) for fyaqa__cac, vsw__jwrzy in
        enumerate(col_typs) if vsw__jwrzy.dtype == types.NPDatetime('ns'))
    nwee__hhxxq = ', '.join(["{}='{}'".format(eye__mdvtn, bodo.ir.csv_ext.
        _get_dtype_str(vsw__jwrzy)) for eye__mdvtn, vsw__jwrzy in zip(
        pkg__mrc, col_typs)])
    iyf__wqjcb = ', '.join(["'{}':{}".format(ptny__vjpw, bodo.ir.csv_ext.
        _get_pd_dtype_str(vsw__jwrzy)) for ptny__vjpw, vsw__jwrzy in zip(
        col_names, col_typs)])
    if compression is None:
        compression = 'uncompressed'
    zkc__mzijf = 'def json_reader_py(fname):\n'
    zkc__mzijf += '  df_typeref_2 = df_typeref\n'
    zkc__mzijf += '  check_java_installation(fname)\n'
    zkc__mzijf += f"""  bucket_region = bodo.io.fs_io.get_s3_bucket_region_njit(fname, parallel={parallel})
"""
    if storage_options is None:
        storage_options = {}
    storage_options['bodo_dummy'] = 'dummy'
    zkc__mzijf += (
        f'  storage_options_py = get_storage_options_pyobject({str(storage_options)})\n'
        )
    zkc__mzijf += (
        '  f_reader = bodo.ir.json_ext.json_file_chunk_reader(bodo.libs.str_ext.unicode_to_utf8(fname), '
        )
    zkc__mzijf += (
        """    {}, {}, -1, bodo.libs.str_ext.unicode_to_utf8('{}'), bodo.libs.str_ext.unicode_to_utf8(bucket_region), storage_options_py )
"""
        .format(lines, parallel, compression))
    zkc__mzijf += '  bodo.utils.utils.check_and_propagate_cpp_exception()\n'
    zkc__mzijf += '  if bodo.utils.utils.is_null_pointer(f_reader):\n'
    zkc__mzijf += "      raise FileNotFoundError('File does not exist')\n"
    zkc__mzijf += f'  with objmode({nwee__hhxxq}):\n'
    zkc__mzijf += f"    df = pd.read_json(f_reader, orient='{orient}',\n"
    zkc__mzijf += f'       convert_dates = {convert_dates}, \n'
    zkc__mzijf += f'       precise_float={precise_float}, \n'
    zkc__mzijf += f'       lines={lines}, \n'
    zkc__mzijf += '       dtype={{{}}},\n'.format(iyf__wqjcb)
    zkc__mzijf += '       )\n'
    zkc__mzijf += (
        '    bodo.ir.connector.cast_float_to_nullable(df, df_typeref_2)\n')
    for eye__mdvtn, ptny__vjpw in zip(pkg__mrc, col_names):
        zkc__mzijf += '    if len(df) > 0:\n'
        zkc__mzijf += "        {} = df['{}'].values\n".format(eye__mdvtn,
            ptny__vjpw)
        zkc__mzijf += '    else:\n'
        zkc__mzijf += '        {} = np.array([])\n'.format(eye__mdvtn)
    zkc__mzijf += '  return ({},)\n'.format(', '.join(lfqnq__gham for
        lfqnq__gham in pkg__mrc))
    dqf__tzzfh = globals()
    dqf__tzzfh.update({'bodo': bodo, 'pd': pd, 'np': np, 'objmode': objmode,
        'check_java_installation': check_java_installation, 'df_typeref':
        bodo.DataFrameType(tuple(col_typs), bodo.RangeIndexType(None),
        tuple(col_names)), 'get_storage_options_pyobject':
        get_storage_options_pyobject})
    nmsn__fam = {}
    exec(zkc__mzijf, dqf__tzzfh, nmsn__fam)
    cyyr__kkfg = nmsn__fam['json_reader_py']
    eks__czlqz = numba.njit(cyyr__kkfg)
    compiled_funcs.append(eks__czlqz)
    return eks__czlqz
